/* Code for retrieving the mp3 files from the SD card using FatFS
   and sending the file to either the STA013 or the VS-1011-LE to play the mp3
   Code adapted from http://frank.circleofcurrent.com/musicclock/ by Frank Zhao,
   which is distributed under the terms of the GNU General Public
   License version 2 or the GNU Lesser General Public License version 2.1

   This file is free software; you can redistribute it and/or modify
   it under the terms of either the GNU General Public License version 2
   or the GNU Lesser General Public License version 2.1, both as
   published by the Free Software Foundation.
*/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdio.h>
#include <avr/power.h>
#include "fatfs/ff.h"
#include "fatfs/diskio.h"
#include "uart.h"
#include "SDcard_audio.h"

FATFS filesystem;
DIR rootDir;
FIL file; //the currently opened file
unsigned char playing; //0 if not playing, 1 if playing
unsigned char file_open; //0 if no file currently open, 1 if a file is open

/* initialize the SD card, mount the FATFs filesystem and open the root directory*/
void init_SD(void) {
	FRESULT errCode = -1;
    fprintf(stdout,("Initializing SD card\r\n"));
    while (errCode != FR_OK) {
        fprintf(stdout,("Mounting File System\r\n"));
        errCode = f_mount(0, &filesystem);
        if (errCode != FR_OK) {
            fprintf(stdout,("Error: f_mount failed, error code: %d\r\n"), errCode);
            _delay_ms(1000); // retry after 1 sec
            continue;
        }
        fprintf(stdout,("Opening Root Directory\r\n"));
        errCode = f_opendir(&rootDir, "/");
        if (errCode != FR_OK) {
            fprintf(stdout,("Error: f_opendir failed, error code: %d\r\n"), errCode);
            _delay_ms(1000); // retry after 1 sec
            continue;
        }
    }
	file_open = 0;
	playing = 0;
}

//play the next MP3 file
int play_next_MP3(void) {
	FRESULT errCode = -1;
	if (file_open == 1) {
		errCode = f_close(&file);
		if (errCode != FR_OK) fprintf(stdout,("Error: f_close failed, error code: %d\r\n"), errCode);
	}
	// read all file entries
    FILINFO fileEntry;
    errCode = f_readdir(&rootDir, &fileEntry);
    if (errCode != FR_OK) {
        fprintf(stdout,("Error: f_readdir failed, error code: %d\r\n"), errCode);
        while (1); // freeze
    }
    if (fileEntry.fname[0] == 0) {
        fprintf(stdout,("End of All Files\r\n"));
		errCode = f_readdir(&rootDir, NULL); //rewind back to start
    } else if (fileEntry.fname[0] == '.') {
        fprintf(stdout,("Skipping \"%s\"\r\n\r\n"), fileEntry.fname);
    } else if (fileEntry.fattrib & AM_DIR) {
        fprintf(stdout,("Directory Entry: %s\r\n\r\n"), fileEntry.fname);
    } else {
        fprintf(stdout,("File Entry: %s\r\n"), fileEntry.fname);            
        errCode = f_open(&file, fileEntry.fname, FA_READ); // open file
        if (errCode != FR_OK) {
            fprintf(stdout,("Error: f_open failed, error code: %d\r\n"), errCode);
            while (1); // freeze
        }
		fprintf(stdout,("start playing: %s\r\n"), fileEntry.fname);
		playing = 1;
		file_open = 1;
    }
	return 0;
}

//play the greeting MP3 file
int play_greeting(void) {
	FRESULT errCode = -1;
    fprintf(stdout,"Opening Greeting File\r\n");
    errCode = f_open(&file, "/greet.mp3", FA_READ); // open file
    if (errCode != FR_OK) {
        fprintf(stdout,("Error: f_open failed, error code: %d\r\n"), errCode);
    }
    playing = 1;
	file_open = 1;
    return 0;
}

//stop playing audio
void stop_playing(void) {
	FRESULT errCode = -1;
	if (file_open == 1) {
		errCode = f_close(&file);
		if (errCode != FR_OK) fprintf(stdout,("Error: f_close failed, error code: %d\r\n"), errCode);
		file_open = 0;
	}
}
