//main file for the display controller

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/twi.h>
#include <avr/power.h>

#include "usbdrv/usbdrv.h"
#include "usbdrv/oddebug.h"
#include "uart.h"
#include "SDcard_audio.h"
#include "fatfs/ff.h"

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

#define t2 20 //time between USB reports
#define t1 100 //time for task
#define PIR_detect (PINA & 1<<PINA4)
#define idle_delay 100 //time until controller goes back to idle state

//timeout counter for task1
volatile unsigned char time1;
volatile unsigned char time2;
unsigned int idle_countdown;

unsigned char controller_state;
unsigned char mp3button_state;
//State machine state names
#define Idle 0
#define MaybePressed 1
#define Detected 1
#define Pressed 2
#define MaybeReleased 3

//MP3 variables and defines
unsigned int mp3count, mp3doNext;

static uchar    reportBuffer[2];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */
unsigned char key; //key to be sent to computer, either left, right or nothing

//USB report descriptor, based on the HIDKeys example at http://www.obdev.at/products/vusb/index.html
PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = { /* USB report descriptor */
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};

//sends a USB report to the computer, called at regular intervals of 40ms by
//timer0 overflow ISR. We have the ISR call it instead of main due to the
//critical timing requirements of sending this report
static void buildReport(void) {
  usbPoll();
  if(usbInterruptIsReady()) {
	reportBuffer[0] = 0;
	reportBuffer[1] = key; //the keystroke that we are sending
	usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
	key = 0x00; //clear the key so that only one keypress is registered until task1 runs again
  }
}

//**********************************************************
//timer 0 overflow ISR
ISR (TIMER0_COMPA_vect) {
  //Decrement the time if not already zero
  if (time1 > 0)	--time1;
  if (time2 > 0)	--time2;
  else {
  	time2=t2;
	//we call this function from the ISR instead of main due to its timing critical nature
	buildReport();
  }
}

//**********************************************************
//sets up the USB function, code taken from the HID keys example provided by http://www.obdev.at/products/vusb/index.html
uchar	usbFunctionSetup(uchar data[8]) {
	usbRequest_t    *rq = (void *)data;
    usbMsgPtr = reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {    /* class request type */
        if (rq->bRequest == USBRQ_HID_GET_REPORT) {  /* Value: ReportType (highbyte), ReportID (lowbyte) */
            /* we only have one report type, so don't look at wValue */
            buildReport();
            return sizeof(reportBuffer);
        } else if(rq->bRequest == USBRQ_HID_GET_IDLE) {
            usbMsgPtr = &idleRate;
            return 1;
        } else if(rq->bRequest == USBRQ_HID_SET_IDLE) {
            idleRate = rq->wValue.bytes[1];
        }
    } else {
        /* no vendor specific requests implemented */
    }
	return 0;
}

//polls the left and right sensors and buttons and sets the keystroke to be sent by USB
int detect_control() {
	if (~PINA & 0b00000101) { //right sensor or right button
		key = 0x4F; //4F == right
		//fprintf(stdout, "right");
		return 1;
	} else if (~PINA & 0b00001010) { //left sensor or left button
		key = 0x50; //50 == left
		//fprintf(stdout, "left");
		return 1;
	}
	return 0;
}

//**********************************************************
//Timed task subroutine
//Task 1 runs a state machine that polls the PIR sensors and buttons
//it then selects and plays sounds such as greeting messages
void task1(void) {
  //fprintf(stdout, "state=%d PINA=%x\r\n", controller_state, PINA);
  if (idle_countdown > 0) idle_countdown--;
  else {
  	controller_state = Idle;
	playing = 0;
	PORTA |= (1<<PORTA7); // turn off LED
	stop_playing();
  }
  switch (controller_state) {
	  //no person detected
	  case Idle:
	  	if (detect_control()) { //poll the top IR sensors
			PORTA &= ~(1<<PORTA7 | 1<<PORTA6); //turn on both LEDs
			controller_state = Pressed;
			play_greeting();
			idle_countdown = idle_delay;
		}
	    else if (PIR_detect) {
			PORTA &= ~(1<<PORTA7);
			controller_state = Detected;
			key = 0x04; //send a keypress of "a" to wake up the computer if it has turned of its display
			play_greeting();
			idle_countdown = idle_delay;
		}
		break;
	  //person detected, no button pressed
	  case Detected:
	  	if (detect_control()) {
			controller_state = Pressed;
			idle_countdown = idle_delay;
			PORTA &= ~(1<<PORTA6);
		}
		if (PIR_detect) idle_countdown = idle_delay;
		break;
	  //left or right button pressed
	  case Pressed:
	  	if ((PINA & 0x0f)== 0b00001111) controller_state = MaybeReleased;
	  	idle_countdown = idle_delay;
	    break;

	  case MaybeReleased:
	  	if ((PINA & 0x0f) == 0b00001111) {
			controller_state = Detected;
			PORTA |= 1<<PORTA6;
		}
		else controller_state = Pressed;
	    break;
  }

  // MP3 Player button State machine
  switch (mp3button_state) {
	  //MP3 button released
	  case Idle:
	  	if (~PINA & (1<<PINA5)) mp3button_state = Pressed;
		break;
	  //MP3 button pressed
	  case Pressed:
	  	if (PINA & (1<<PINA5)) mp3button_state = MaybeReleased;
		//continue song if button was held for 2s or longer.
		//otherwise skip to the next song as usual.
	  	if (mp3count == 20) play_next_MP3();
		if (mp3count <= 20) ++mp3count;
	    break;

	  case MaybeReleased:
	  	if (PINA & (1<<PINA5)) {
			mp3button_state = Idle;
			//if we are playing, stop playing
			if (playing == 1 && mp3count < 20) playing = 0;
			//if we are stopped, resume play
			else playing = 1;
			
			//reset defaults
			mp3count = 0;
		}
		else mp3button_state = Pressed;
	    break;
  }
}


//**********************************************************
//Initialize the timers, connect to the computer via USB and set up the state varaibles
void initialize(void) {
  //set up timer 0 for 1 mSec timebase
  TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR
  OCR0A = 249;  		//set the compare re to 250 time ticks
  //set prescalar to divide by 64
  TCCR0B= 3; //0b00001011;
  // turn on clear-on-match
  TCCR0A= (1<<WGM01) ;

    //init the UART -- uart_init() is in uart.c
  uart_init();
  stdout = stdin = stderr = &uart_str;
  fprintf(stdout,"Initializing...\n\r");

  DDRA = 1<<DDA7 | 1<<DDA6; //two LEDs
  //one LED turns on when a person is detected
  //the other turns on when a button is pressed
  PORTA |= 1<<PORTA7 | 1<<PORTA6; //turn off both LEDs

  //init the task timer
  time1=t1;
  time2=t2;

  //USB initialization
  usbInit();
  usbDeviceDisconnect();
  _delay_ms(500);
  usbDeviceConnect();

  //initialize the state variables
  idle_countdown = 0;
  mp3count = 0;
  controller_state = Idle;
  key = 0;

  //MP3 and SD card port initializations
  SPI_PORTOUTPUT = 0xff;
  SPI_PORTDIR = 0xff;
  SPI_PORTDIR &= ~_BV(SPI_PIN_MISO);
  SPI_PORTDIR &= ~_BV(MP3_PIN_DREQ);
  MP3_PORTOUTPUT &= ~_BV(MP3_PIN_RST); // hold VS1011 in power-down mode until init_VS

  //crank up the ISRs
  sei();
}

//**********************************************************
//Entry point and task scheduler loop
int main(void) {
  mp3_playStatus_t playstatus; //VS_playfile returns PLAYSTATUS_ENDOFFILE if it has finished current file
  initialize();
  //init_STA013(); //initializes the STA013
  init_SD();
  init_VS();
  // main task scheduler loop
  while(1) {
  	if (time1 == 0) {time1 = t1; task1(); }
	if (playing == 1) {
		playstatus = VS_playfile(&file);
		if (playstatus == PLAYSTATUS_ENDOFFILE) { //finished playing the current MP3 file
			play_next_MP3();
		}
	}
  }
}
